"""A package for already-implemented machine learning Siamese architectures.
"""

from dualing.models.contrastive import ContrastiveSiamese
from dualing.models.cross_entropy import CrossEntropySiamese
from dualing.models.triplet import TripletSiamese
